
//  @(#)print.C 1.6 93/12/22
//
//  Copyright (c) Steve Holden and Rich Burridge.
//                All rights reserved.
//
//  Permission is given to distribute these sources, as long as the
//  copyright messages are not removed, and no monies are exchanged.
//
//  No responsibility is taken for any errors inherent either
//  to the comments or the code of this program, but if reported
//  to me then an attempt will be made to fix them.

#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include "option.h"
#include "header.h"
#include "input.h"
#include "print.h"

extern Header hdr ; 
extern Input  input ;  
extern Option opt ;
extern Print  prt ;


void Print::
do_date(void)     // Output Postscript definition for the date and time.
{
  char *ptr ;            // Pointer to current time or date string.
  int len ;
  time_t clock ;         // Used by the localtime function call.
  struct stat status ;   // Stat(2) information from current file (-modtime).
  struct tm *tm ;        // Used by the localtime and asctime calls.
  char *timenow ;        // Used to set TimeNow field with users name.

  if (opt.modtime && opt.numnames)         // -modtime given on command line?
    if (!stat(prt.curfname, &status))
      {
        tm = localtime(&status.st_mtime) ;
        hdr.date = asctime(tm) ;
        hdr.date[24] = '\0' ;
      }

  if (hdr.date == NULL)
    {
      clock = time((time_t *) 0) ;
      tm = localtime(&clock) ;
      ptr = asctime(tm) ;
      ptr[24] = '\0' ;
    }
  else ptr = hdr.date ;

  if (!opt.article) psdef("TimeNow", ptr) ;
  else
    { 
      len = strlen(ptr) ;
      timenow = (char *)
         malloc((unsigned) (len + 6 + strlen(input.whoami) + 1)) ;
      SPRINTF(timenow, "%s  - (%s)", ptr, input.whoami) ;
      psdef("TimeNow", timenow) ;
    }
}


void Print::
end(enum action_type op)
{
  int cval, pval ;

  switch (op)
    {
      case IS_COL  : FPUTS("sf ", ofp) ;
                     linect = 1 ;
                     ++colct ;
                     if (opt.landscape && !opt.forcepage)
                       pval = (2 * pageno) + colct - 2 ;
                     else
                       pval = pageno ;
                     if (opt.forcepage) cval = 1 ;
                     else                 cval = colct ;
                     FPRINTF(ofp, "(%1d) %1d endcol\n", pval, cval) ;

// Needed for correct heading in multiple columns.

                     if (opt.numcols > 1) set_defs() ;
                     if (opt.landscape && !opt.forcepage &&
                         colct != opt.numcols) return ;
                     pageno++ ;
                     break ;
      case IS_FILE : linect = 0 ;
                     break ;
      case IS_LINE : FPUTS(") showline\n", ofp) ;
                     break ;
      case IS_PAGE : linect = 0 ;        
                     FPRINTF(ofp, "(%1d) endpage\n", pageno-1) ;
    }
}


// Display a string with PostScript-sensitive characters escaped.

void Print::
expand(unsigned char *s)
{
  for (; s && *s; s++)
    {
      switch (*s)
        {
          case '\\' : FPUTS("\\\\", ofp) ;
                      break ;
          case '('  : FPUTS("\\(", ofp) ;
                      break ;
          case ')'  : FPUTS("\\)", ofp) ;
                      break ;
          default   : if ((*s < 0x20) || (*s > 0x7e))
                        FPRINTF(ofp, "\\%03o", *s) ;
                      else
                        if (isprint(*s)) PUTC(*s, ofp) ;
        }
    }
}


FILE *Print::
makecmd(char *fname)
{
  FILE *fp ;              // Output stream pointer from popen().
  char cmd[MAXLINE] ;     // The print command to execute.

  if (ofp && opt.toprinter) PCLOSE(ofp) ;
  STRCPY(cmd, (opt.bsdprint ? "lpr -J " : "lp -t ")) ;
  STRCAT(cmd, fname) ;
  STRCAT(cmd, " ") ;
  if (!opt.banner) STRCAT(cmd, (opt.bsdprint) ? "-h " : "-o nobanner ") ;

// If no printer name given, try to get the default from the environment.

  if (!opt.printer_name)
    {
      if (opt.bsdprint) opt.printer_name = getenv("PRINTER") ;
      else              opt.printer_name = getenv("LPDEST") ;
    }
  if (opt.printer_name)
    {
      STRCAT(cmd, (opt.bsdprint ? "-P " : "-d ")) ;
      STRCAT(cmd, opt.printer_name) ;
    }
  if ((fp = popen(cmd, "w")) == NULL)
    {
      cerr << "Invalid print command: " << cmd << "\n" ;
      cerr << "Defaulting to standard output.\n" ;
      fp = stdout ;
    }
  return(fp) ;
}


Print::
Print()
{
  colct        = 0 ;              // Column count on current page.
  linect       = 0 ;              // Line count on current page.
  ofp          = NULL ;           // No output stream open initially.
  pageno       = 1 ;              // Page number within message.
  tpn          = 0 ;              // Total number of pages printed.
}


void Print::
print_extra(void)
{
  int i ;
 
  for (i = 0; i < opt.addhdr_num; i++)
    if (input.hdrs[i])
      prt.show_text(T_MIXED, opt.addhdrs[i], input.hdrs[i]) ;
}


void Print::
psdef(char *name, char *def)       // Do a PostScript define.
{
  FPRINTF(ofp, "/%s (", name) ;
  expand((unsigned char *) def) ;
  FPUTS(") def\n", ofp) ;
}


void Print::
psdict(char *name, char *state)    // Write dictionary definition.
{
  FPRINTF(ofp, "statusdict /%s known { statusdict begin %s %s end } if\n",
          name, state, name) ;
}


void Print::
set_defs(void)                     // Setup PostScript definitions.
{
  int i ;

       if (opt.article)                message_for = "Article from" ;
  else if (opt.print_orig && hdr.from) message_for = "From" ;
  psdef("MailFor", message_for) ;

  if (opt.article && hdr.newsgroups)          // User.
    {
      for (i = 0; i < strlen(hdr.newsgroups); i++)
        if (hdr.newsgroups[i] == ',' ||
            hdr.newsgroups[i] == '\0') break ;
      input.owner = (char *) realloc(input.owner, (unsigned int) i+1) ;
      STRNCPY(input.owner, hdr.newsgroups, i) ;
      input.owner[i] = '\0' ;
    }
  else if (opt.print_orig && hdr.from)
    {
      i = strlen(hdr.from) ;
      input.owner = (char *) realloc(input.owner, (unsigned int) i+1) ;
      STRNCPY(input.owner, hdr.from, i) ;
      input.owner[i] = '\0' ;
    }

  psdef("User", input.owner) ;
 
  do_date() ;                                                    // TimeNow.
 
  if (opt.text_doc && opt.numnames) hdr.subject = curfname ;
  psdef("Subject", (opt.gsubject != NULL) ? opt.gsubject
                                          : hdr.subject) ;       // Subject.
}


// Display the PostScript prologue file for mp.
                   
void Print::
show_prologue(char *pro)
{
  FILE *cf, *pf ;
  char buf[MAXLINE], tmpstr[MAXLINE] ;
  char cpro[MAXPATHLEN] ;     // Full pathname of the common prologue file.
  int t2 ;                    // Possible extract page or line length.

  if ((pf = fopen(pro, "r")) == NULL)                                
    {
      SPRINTF(tmpstr, "%s/%s", opt.prologue, pro) ;
      if ((pf = fopen(tmpstr, "r")) == NULL)
        {
          cerr << opt.appname << ": Prologue file " << pro <<
                  " not found.\n" ;
          exit(1) ;
        }
    }

  while (fgets(buf, MAXLINE, pf) != NULL)
    {

// If "-columns #", "-linelength #" or "-pagelength #" was given on the
// command line, then use that value, otherwise get the appropriate value
// from the prologue file.

           if (opt.plen_given && EQUALN(buf, "%%PageLength"))
        SPRINTF(buf, "%%%%PageLength %1d\n", opt.plen) ;
      else if (opt.llen_given && EQUALN(buf, "%%LineLength"))
        SPRINTF(buf, "%%%%LineLength %1d\n", opt.llen) ;
      else if (opt.cols_given && EQUALN(buf, "%%NumCols"))
        SPRINTF(buf, "%%%%NumCols %1d\n",    opt.numcols) ;
      FPUTS(buf, ofp) ;

// If appropriate, extract out the pagelength, linelength or the number of
// columns. Note that they might have just been set above. It was simpler to
// do it this way.
 
      tmpstr[0] = '\0' ;
      SSCANF(buf, "%s %d", tmpstr, &t2) ;
           if (EQUAL(tmpstr, "%%PageLength"))
        opt.plen = t2 ;                  // Change the page length. */
      else if (EQUAL(tmpstr, "%%LineLength"))
        opt.llen = t2 ;                  // Change the line length. */
      else if (EQUAL(tmpstr, "%%NumCols"))
        opt.numcols = t2 ;               // Change the number of columns. */
      else if (EQUAL(tmpstr, "%%EndComments"))
        {

//  If this is the %%EndComments line from the prologue file, then we
//  need to read (and output to stdout), the contents of the common
//  prologue file, mp.common.ps

          SPRINTF(cpro, "%s/mp.common.ps", opt.prologue) ;
          if ((cf = fopen(cpro, "r")) == NULL)
            {
              cerr << opt.appname << ": Common prologue file " << cpro <<
                   " not found.\n" ;
              exit(1) ;
            }
          while (fgets(buf, MAXLINE, cf) != NULL) FPUTS(buf, ofp) ;
          FCLOSE(cf) ;

// Output the sizes of the four paper margins.
 
          SPRINTF(buf, "/bottomOffset %1d def\n", opt.bmargin) ;
          FPUTS(buf, ofp) ;
          SPRINTF(buf, "/leftOffset %1d def\n", opt.lmargin) ;
          FPUTS(buf, ofp) ;
          SPRINTF(buf, "/rightOffset %1d def\n", opt.rmargin) ;
          FPUTS(buf, ofp) ;
          SPRINTF(buf, "/topOffset %1d def\n", opt.tmargin) ;
          FPUTS(buf, ofp) ;

#ifdef WANTED
//  Adjust for US or A4 paper size if needed.

               if (opt.paper_size == P_US)
            {
              FPUTS("/fullwidth 8.5 inch def\n", ofp) ;
              FPUTS("/fullheight 11 inch def\n", ofp) ;
            }
          else if (opt.paper_size == P_A4)
            {
              FPUTS("/fullwidth  595 def\n", ofp) ;
              FPUTS("/fullheight 842 def\n", ofp) ;
            }
#endif /*WANTED*/
        }
    }
  FCLOSE(pf) ;

  psdict("setduplexmode", (opt.duplex) ? "true" : "false") ;
  if (opt.tumble) psdict("settumble",     "true") ;
  if (opt.flip)   FPUTS("switchtopbottom\n", ofp) ;
  FPUTS("%%EndProlog\n", ofp) ;
}


void Print::
show_text(enum text_type op, char *hdr, char *str)
{
  static char *fontstr[MAXTTYPES] = {
                 "BoldFont ",              // T_BOLD.
                 "BoldFont ",              // T_MIXED.
                 NULL,                     // T_PLAIN.
                 "pf "                     // T_ROMAN.
  } ;

  useline() ;
  if (fontstr[(int) op]) FPUTS(fontstr[(int) op], ofp) ;

  start(IS_LINE) ;
  switch (op)
    {
      case T_BOLD  : expand((unsigned char *) hdr) ;
                     expand((unsigned char *) str) ;
                     break ;
      case T_MIXED : expand((unsigned char *) hdr) ;
                     FPUTS(") show pf (", ofp) ;
                     expand((unsigned char *) str) ;
                     break ;
      case T_PLAIN : expand((unsigned char *) str) ;
                     break ;
      case T_ROMAN : expand((unsigned char *) str) ;
    }
  end(IS_LINE) ;
}


void Print::
show_trailer(void)
{
  FPUTS("%%Trailer\n", ofp) ;
  FPRINTF(ofp, "%%%%Pages: %1d\n", prt.tpn) ;
}


void Print::
start(enum action_type op)
{
  switch (op)
    {
      case IS_COL  : break ;
      case IS_FILE : break ;
      case IS_LINE : PUTC('(', ofp) ;
                     break ;
      case IS_PAGE : FPRINTF(ofp, "%%%%Page: ? %1d\n", ++tpn) ;
                     FPRINTF(ofp, "%1d newpage\n", pageno) ;
                     set_defs() ;
                     FPUTS("sf ", ofp) ;
                     break ;
    }
}


void Print::
useline(void)      // Called in order to ready a line for printing.
{
  if (++linect > opt.plen || input.end_of_page)
    {
      end(IS_COL) ;
      if (input.end_of_page && opt.forcepage) colct = opt.numcols ;
      if (colct < opt.numcols) return ;
      colct = 0 ;
      end(IS_PAGE) ;
      linect = 1 ;
      start(IS_PAGE) ;
    }
}
